/*
 * kgraft_patch_bsc1004419
 *
 * Fix for CVE-2016-5195, bsc#1004419
 *
 *  Upstream commit:
 *  19be0eaffa3a ("mm: remove gup_flags FOLL_WRITE games from __get_user_pages()")
 *
 *  SLE12-SP1 commit:
 *  e91ab14461ccffe716d0bc9d22c7eb9b21409767
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if defined(CONFIG_X86_64)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/mm_types.h>
#include <linux/hugetlb.h>
#include <linux/rmap.h>
#include <linux/swap.h>
#include <linux/swapops.h>
#include <linux/pagemap.h>

static void (*kgr_mlock_vma_page)(struct page *page);
static void (*kgr_lru_add_drain)(void);
static struct page *(*kgr_vm_normal_page)(struct vm_area_struct *vma, unsigned long addr, pte_t pte);
static unsigned long *kgr_zero_pfn;
static void (*kgr_migration_entry_wait)(struct mm_struct *mm, pmd_t *pmd, unsigned long address);
static struct page *(*kgr_follow_trans_huge_pmd)(struct vm_area_struct *vma, unsigned long addr, pmd_t *pmd, unsigned int flags);
static struct mm_struct *kgr_init_mm;
static int (*kgr_in_gate_area)(struct mm_struct *mm, unsigned long addr);
static struct vm_area_struct *(*kgr_get_gate_vma)(struct mm_struct *mm);
static int (*kgr_handle_mm_fault)(struct mm_struct *mm, struct vm_area_struct *vma, unsigned long address, unsigned int flags);
static struct vm_area_struct *(*kgr_find_extend_vma)(struct mm_struct * mm, unsigned long addr);

#if !defined(CONFIG_XEN)
static struct page *(*kgr_follow_huge_pud)(struct mm_struct *mm, unsigned long address, pud_t *pud, int flags);
static int (*kgr_pmd_huge)(pmd_t pmd);
static struct page *(*kgr_follow_huge_addr)(struct mm_struct *mm, unsigned long address, int write);
static int (*kgr_pud_huge)(pud_t pud);
static struct page *(*kgr_follow_huge_pmd)(struct mm_struct *mm, unsigned long address, pmd_t *pmd, int flags);
static void (*kgr__split_huge_page_pmd)(struct vm_area_struct *vma, unsigned long address, pmd_t *pmd);
static long (*kgr_follow_hugetlb_page)(struct mm_struct *mm, struct vm_area_struct *vma, struct page **pages, struct vm_area_struct **vmas, unsigned long *position, unsigned long *nr_pages, long i, unsigned int flags);

#define kgr_split_huge_page_pmd(__vma, __address, __pmd)		\
	do {								\
		pmd_t *____pmd = (__pmd);				\
		if (unlikely(pmd_trans_huge(*____pmd)))			\
			kgr__split_huge_page_pmd(__vma, __address,	\
					____pmd);			\
	}  while (0)

#else

#define kgr_follow_huge_pud(mm, addr, pud, flags)	NULL
#define kgr_pmd_huge(x)	0
#define kgr_follow_huge_addr(mm, addr, write)	ERR_PTR(-EINVAL)
#define kgr_pud_huge(x)	0
#define kgr_follow_huge_pmd(mm, addr, pmd, flags)	NULL
#define kgr_follow_hugetlb_page(m,v,p,vs,a,b,i,w)	({ BUG(); 0; })

#define kgr_split_huge_page_pmd(__vma, __address, __pmd)	\
	do { } while (0)

#endif /* !CONFIG_XEN */

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
#if !defined(CONFIG_XEN)
	{ "follow_huge_pud", (void *)&kgr_follow_huge_pud },
	{ "pmd_huge", (void *)&kgr_pmd_huge },
	{ "follow_huge_addr", (void *)&kgr_follow_huge_addr },
	{ "pud_huge", (void *)&kgr_pud_huge },
	{ "follow_huge_pmd", (void *)&kgr_follow_huge_pmd },
	{ "__split_huge_page_pmd", (void *)&kgr__split_huge_page_pmd },
	{ "follow_hugetlb_page", (void *)&kgr_follow_hugetlb_page },
	{ "follow_trans_huge_pmd", (void *)&kgr_follow_trans_huge_pmd },
#endif
	{ "mlock_vma_page", (void *)&kgr_mlock_vma_page },
	{ "lru_add_drain", (void *)&kgr_lru_add_drain },
	{ "vm_normal_page", (void *)&kgr_vm_normal_page },
	{ "zero_pfn", (void *)&kgr_zero_pfn },
	{ "migration_entry_wait", (void *)&kgr_migration_entry_wait },
	{ "init_mm", (void *)&kgr_init_mm },
	{ "in_gate_area", (void *)&kgr_in_gate_area },
	{ "get_gate_vma", (void *)&kgr_get_gate_vma },
	{ "handle_mm_fault", (void *)&kgr_handle_mm_fault },
	{ "find_extend_vma", (void *)&kgr_find_extend_vma },
};

#define kgr_pgd_offset_k(address) pgd_offset(kgr_init_mm, (address))

/*
 * Fix CVE-2016-5195
 *  +next line
 */
#define FOLL_COW       0x4000  /* internal GUP flag */

static inline int kgr_is_zero_pfn(unsigned long pfn)
{
	extern unsigned long zero_pfn;
	return pfn == *kgr_zero_pfn;
}

/* inlined */
static inline void kgr__get_page_tail_foll(struct page *page,
					bool get_page_head)
{
	/*
	 * If we're getting a tail page, the elevated page->_count is
	 * required only in the head page and we will elevate the head
	 * page->_count and tail page->_mapcount.
	 *
	 * We elevate page_tail->_mapcount for tail pages to force
	 * page_tail->_count to be zero at all times to avoid getting
	 * false positives from get_page_unless_zero() with
	 * speculative page access (like in
	 * page_cache_get_speculative()) on tail pages.
	 */
	VM_BUG_ON(atomic_read(&page->first_page->_count) <= 0);
	VM_BUG_ON(atomic_read(&page->_count) != 0);
	VM_BUG_ON(page_mapcount(page) < 0);
	if (get_page_head)
		atomic_inc(&page->first_page->_count);
	if (compound_tail_refcounted(page->first_page))
		atomic_inc(&page->_mapcount);
}

/* inlined */
static inline void kgr_get_page_foll(struct page *page)
{
	if (unlikely(PageTail(page)))
		/*
		 * This is safe only because
		 * __split_huge_page_refcount() can't run under
		 * get_page_foll() because we hold the proper PT lock.
		 */
		kgr__get_page_tail_foll(page, true);
	else {
		/*
		 * Getting a normal page or the head of a compound page
		 * requires to already have an elevated page->_count.
		 */
		VM_BUG_ON(atomic_read(&page->_count) <= 0);
		atomic_inc(&page->_count);
	}
}

/*
 * Fix CVE-2016-5195
 * +new function
 */
static inline bool kgr_can_follow_write_pte(pte_t pte, unsigned int flags)
{
	return pte_write(pte) ||
		((flags & FOLL_FORCE) && (flags & FOLL_COW) && pte_dirty(pte));
}

/* patched function */
struct page *kgr_follow_page_mask(struct vm_area_struct *vma,
			      unsigned long address, unsigned int flags,
			      unsigned int *page_mask)
{
	pgd_t *pgd;
	pud_t *pud;
	pmd_t *pmd;
	pte_t *ptep, pte;
	spinlock_t *ptl;
	struct page *page;
	struct mm_struct *mm = vma->vm_mm;

	*page_mask = 0;

	page = kgr_follow_huge_addr(mm, address, flags & FOLL_WRITE);
	if (!IS_ERR(page)) {
		BUG_ON(flags & FOLL_GET);
		goto out;
	}

	page = NULL;
	pgd = pgd_offset(mm, address);
	if (pgd_none(*pgd) || unlikely(pgd_bad(*pgd)))
		goto no_page_table;

	pud = pud_offset(pgd, address);
	if (pud_none(*pud))
		goto no_page_table;
	if (kgr_pud_huge(*pud) && vma->vm_flags & VM_HUGETLB) {
		page = kgr_follow_huge_pud(mm, address, pud, flags);
		if (page)
			goto out;
		goto no_page_table;
	}
	if (unlikely(pud_bad(*pud)))
		goto no_page_table;

	pmd = pmd_offset(pud, address);
	if (pmd_none(*pmd))
		goto no_page_table;
	if (kgr_pmd_huge(*pmd) && vma->vm_flags & VM_HUGETLB) {
		page = kgr_follow_huge_pmd(mm, address, pmd, flags);
		if (page)
			goto out;
		goto no_page_table;
	}
	if ((flags & FOLL_NUMA) && pmd_protnone(*pmd))
		goto no_page_table;
	if (pmd_trans_huge(*pmd)) {
		if (flags & FOLL_SPLIT) {
			kgr_split_huge_page_pmd(vma, address, pmd);
			goto split_fallthrough;
		}
		ptl = pmd_lock(mm, pmd);
		if (likely(pmd_trans_huge(*pmd))) {
			if (unlikely(pmd_trans_splitting(*pmd))) {
				spin_unlock(ptl);
				wait_split_huge_page(vma->anon_vma, pmd);
			} else {
				page = kgr_follow_trans_huge_pmd(vma, address,
							     pmd, flags);
				spin_unlock(ptl);
				*page_mask = HPAGE_PMD_NR - 1;
				goto out;
			}
		} else
			spin_unlock(ptl);
		/* fall through */
	}
split_fallthrough:
	if (unlikely(pmd_bad(*pmd)))
		goto no_page_table;

	ptep = pte_offset_map_lock(mm, pmd, address, &ptl);

	pte = *ptep;
	if (!pte_present(pte)) {
		swp_entry_t entry;
		/*
		 * KSM's break_ksm() relies upon recognizing a ksm page
		 * even while it is being migrated, so for that case we
		 * need migration_entry_wait().
		 */
		if (likely(!(flags & FOLL_MIGRATION)))
			goto no_page;
		if (pte_none(pte) || pte_file(pte))
			goto no_page;
		entry = pte_to_swp_entry(pte);
		if (!is_migration_entry(entry))
			goto no_page;
		pte_unmap_unlock(ptep, ptl);
		kgr_migration_entry_wait(mm, pmd, address);
		goto split_fallthrough;
	}
	if ((flags & FOLL_NUMA) && pte_protnone(pte))
		goto no_page;
	/*
	 * Fix CVE-2016-5195
	 *  -if ((flags & FOLL_WRITE) && !pte_write(pte))
	 *  +next line
	 */
	if ((flags & FOLL_WRITE) && !kgr_can_follow_write_pte(pte, flags))
		goto unlock;

	page = kgr_vm_normal_page(vma, address, pte);
	if (unlikely(!page)) {
		if ((flags & FOLL_DUMP) ||
		    !kgr_is_zero_pfn(pte_pfn(pte)))
			goto bad_page;
		page = pte_page(pte);
	}

	if (flags & FOLL_GET)
		kgr_get_page_foll(page);
	if (flags & FOLL_TOUCH) {
		if ((flags & FOLL_WRITE) &&
		    !pte_dirty(pte) && !PageDirty(page))
			set_page_dirty(page);
		/*
		 * pte_mkyoung() would be more correct here, but atomic care
		 * is needed to avoid losing the dirty bit: it is easier to use
		 * mark_page_accessed().
		 */
		mark_page_accessed(page);
	}
	if ((flags & FOLL_MLOCK) && (vma->vm_flags & VM_LOCKED)) {
		/*
		 * The preliminary mapping check is mainly to avoid the
		 * pointless overhead of lock_page on the ZERO_PAGE
		 * which might bounce very badly if there is contention.
		 *
		 * If the page is already locked, we don't need to
		 * handle it now - vmscan will handle it later if and
		 * when it attempts to reclaim the page.
		 */
		if (page->mapping && trylock_page(page)) {
			kgr_lru_add_drain();  /* push cached pages to LRU */
			/*
			 * Because we lock page here, and migration is
			 * blocked by the pte's page reference, and we
			 * know the page is still mapped, we don't even
			 * need to check for file-cache page truncation.
			 */
			kgr_mlock_vma_page(page);
			unlock_page(page);
		}
	}
unlock:
	pte_unmap_unlock(ptep, ptl);
out:
	return page;

bad_page:
	pte_unmap_unlock(ptep, ptl);
	return ERR_PTR(-EFAULT);

no_page:
	pte_unmap_unlock(ptep, ptl);
	if (!pte_none(pte))
		return page;

no_page_table:
	/*
	 * When core dumping an enormous anonymous area that nobody
	 * has touched so far, we don't want to allocate unnecessary pages or
	 * page tables.  Return error instead of NULL to skip handle_mm_fault,
	 * then get_dump_page() will return NULL to leave a hole in the dump.
	 * But we can only make this optimization where a hole would surely
	 * be zero-filled if handle_mm_fault() actually did handle it.
	 */
	if ((flags & FOLL_DUMP) &&
	    (!vma->vm_ops || !vma->vm_ops->fault))
		return ERR_PTR(-EFAULT);
	return page;
}

/* inlined */
static inline int kgr_stack_guard_page(struct vm_area_struct *vma, unsigned long addr)
{
	return stack_guard_page_start(vma, addr) ||
	       stack_guard_page_end(vma, addr+PAGE_SIZE);
}

/* patched function */
long kgr__get_user_pages(struct task_struct *tsk, struct mm_struct *mm,
		unsigned long start, unsigned long nr_pages,
		unsigned int gup_flags, struct page **pages,
		struct vm_area_struct **vmas, int *nonblocking)
{
	long i;
	unsigned long vm_flags;
	unsigned int page_mask;

	if (!nr_pages)
		return 0;

	/* 
	 * Require read or write permissions.
	 * If FOLL_FORCE is set, we only require the "MAY" flags.
	 */
	vm_flags  = (gup_flags & FOLL_WRITE) ?
			(VM_WRITE | VM_MAYWRITE) : (VM_READ | VM_MAYREAD);
	vm_flags &= (gup_flags & FOLL_FORCE) ?
			(VM_MAYREAD | VM_MAYWRITE) : (VM_READ | VM_WRITE);

	VM_BUG_ON(!!pages != !!(gup_flags & FOLL_GET));

	/*
	 * If FOLL_FORCE is set then do not force a full fault as the hinting
	 * fault information is unrelated to the reference behaviour of a task
	 * using the address space
	 */
	if (!(gup_flags & FOLL_FORCE))
		gup_flags |= FOLL_NUMA;

	i = 0;

	do {
		struct vm_area_struct *vma;

		vma = kgr_find_extend_vma(mm, start);
		if (!vma && kgr_in_gate_area(mm, start)) {
			unsigned long pg = start & PAGE_MASK;
			pgd_t *pgd;
			pud_t *pud;
			pmd_t *pmd;
			pte_t *pte;

			/* user gate pages are read-only */
			if (gup_flags & FOLL_WRITE)
				return i ? : -EFAULT;
			if (pg > TASK_SIZE)
				pgd = kgr_pgd_offset_k(pg);
			else
				pgd = pgd_offset_gate(mm, pg);
			BUG_ON(pgd_none(*pgd));
			pud = pud_offset(pgd, pg);
			BUG_ON(pud_none(*pud));
			pmd = pmd_offset(pud, pg);
			if (pmd_none(*pmd))
				return i ? : -EFAULT;
			VM_BUG_ON(pmd_trans_huge(*pmd));
			pte = pte_offset_map(pmd, pg);
			if (pte_none(*pte)) {
				pte_unmap(pte);
				return i ? : -EFAULT;
			}
			vma = kgr_get_gate_vma(mm);
			if (pages) {
				struct page *page;

				page = kgr_vm_normal_page(vma, start, *pte);
				if (!page) {
					if (!(gup_flags & FOLL_DUMP) &&
					     kgr_is_zero_pfn(pte_pfn(*pte)))
						page = pte_page(*pte);
					else {
						pte_unmap(pte);
						return i ? : -EFAULT;
					}
				}
				pages[i] = page;
				get_page(page);
			}
			pte_unmap(pte);
			page_mask = 0;
			goto next_page;
		}

#ifdef CONFIG_XEN
		if (vma && (vma->vm_flags & VM_FOREIGN)) {
			struct vm_foreign_map *foreign_map =
				vma->vm_private_data;
			struct page **map = foreign_map->map;
			int offset = (start - vma->vm_start) >> PAGE_SHIFT;
			if (map[offset] != NULL) {
			        if (pages) {
			                struct page *page = map[offset];

					pages[i] = page;
					get_page(page);
				}
				if (vmas)
					vmas[i] = vma;
				i++;
				start += PAGE_SIZE;
				nr_pages--;
				continue;
			}
		}
#endif
		if (!vma ||
		    (vma->vm_flags & (VM_IO | VM_PFNMAP)) ||
		    !(vm_flags & vma->vm_flags))
			return i ? : -EFAULT;

		if (is_vm_hugetlb_page(vma)) {
			i = kgr_follow_hugetlb_page(mm, vma, pages, vmas,
					&start, &nr_pages, i, gup_flags);
			continue;
		}

		do {
			struct page *page;
			unsigned int foll_flags = gup_flags;
			unsigned int page_increm;

			/*
			 * If we have a pending SIGKILL, don't keep faulting
			 * pages and potentially allocating memory.
			 */
			if (unlikely(fatal_signal_pending(current)))
				return i ? i : -ERESTARTSYS;

			cond_resched();
			while (!(page = kgr_follow_page_mask(vma, start,
						foll_flags, &page_mask))) {
				int ret;
				unsigned int fault_flags = 0;

				/* For mlock, just skip the stack guard page. */
				if (foll_flags & FOLL_MLOCK) {
					if (kgr_stack_guard_page(vma, start))
						goto next_page;
				}
				if (foll_flags & FOLL_WRITE)
					fault_flags |= FAULT_FLAG_WRITE;
				if (nonblocking)
					fault_flags |= FAULT_FLAG_ALLOW_RETRY;
				if (foll_flags & FOLL_NOWAIT)
					fault_flags |= (FAULT_FLAG_ALLOW_RETRY | FAULT_FLAG_RETRY_NOWAIT);

				ret = kgr_handle_mm_fault(mm, vma, start,
							fault_flags);

				if (ret & VM_FAULT_ERROR) {
					if (ret & VM_FAULT_OOM)
						return i ? i : -ENOMEM;
					if (ret & (VM_FAULT_HWPOISON |
						   VM_FAULT_HWPOISON_LARGE)) {
						if (i)
							return i;
						else if (gup_flags & FOLL_HWPOISON)
							return -EHWPOISON;
						else
							return -EFAULT;
					}
					if (ret & (VM_FAULT_SIGBUS |
							VM_FAULT_SIGSEGV))
						return i ? i : -EFAULT;
					BUG();
				}

				if (tsk) {
					if (ret & VM_FAULT_MAJOR)
						tsk->maj_flt++;
					else
						tsk->min_flt++;
				}

				if (ret & VM_FAULT_RETRY) {
					if (nonblocking)
						*nonblocking = 0;
					return i;
				}

				/*
				 * The VM_FAULT_WRITE bit tells us that
				 * do_wp_page has broken COW when necessary,
				 * even if maybe_mkwrite decided not to set
				 * pte_write. We can thus safely do subsequent
				 * page lookups as if they were reads. But only
				 * do so when looping for pte_write is futile:
				 * in some cases userspace may also be wanting
				 * to write to the gotten user page, which a
				 * read fault here might prevent (a readonly
				 * page might get reCOWed by userspace write).
				 */
				if ((ret & VM_FAULT_WRITE) &&
				    !(vma->vm_flags & VM_WRITE))
					/*
					 * Fix CVE-2016-5195
					 *  -foll_flags &= ~FOLL_WRITE;
					 *  +next line
					 */
					foll_flags |= FOLL_COW;

				cond_resched();
			}
			if (IS_ERR(page))
				return i ? i : PTR_ERR(page);
			if (pages) {
				pages[i] = page;

				flush_anon_page(vma, page, start);
				flush_dcache_page(page);
				page_mask = 0;
			}
next_page:
			if (vmas) {
				vmas[i] = vma;
				page_mask = 0;
			}
			page_increm = 1 + (~(start >> PAGE_SHIFT) & page_mask);
			if (page_increm > nr_pages)
				page_increm = nr_pages;
			i += page_increm;
			start += page_increm * PAGE_SIZE;
			nr_pages -= page_increm;
		} while (nr_pages && start < vma->vm_end);
	} while (nr_pages);
	return i;
}

static int kgr_patch_bsc1004419_kallsyms(void)
{
        unsigned long addr;
        int i;

        for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
                addr = kallsyms_lookup_name(kgr_funcs[i].name);
                if (!addr) {
                        pr_err("kgraft-patch: symbol %s not resolved\n",
                                kgr_funcs[i].name);
                        return -ENOENT;
                }

                *(kgr_funcs[i].addr) = (void *)addr;
        }

        return 0;
}

int kgr_patch_bsc1004419_init(void)
{
        return kgr_patch_bsc1004419_kallsyms();
}

#endif /* CONFIG_X86_64 */
